@@ -1,22 +1,39 @@
 /*
  * The GGZ Gaming Zone Metaserver Project
- * Copyright (C) 2001 - 2003 Josef Spillner, josef@ggzgamingzone.org
+ * Copyright (C) 2001 - 2004 Josef Spillner, josef@ggzgamingzone.org
  * Published under GNU GPL conditions.
  */
 
+#ifdef HAVE_CONFIG_H
+#  include <config.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <netdb.h>
+#ifdef HAVE_SOCKET_H
+#  include <socket.h>
+#endif
+#ifdef HAVE_NETINET_IN_H
+#  include <netinet/in.h>
+#endif
+#ifdef HAVE_ARPA_INET_H
+#  include <arpa/inet.h>
+#endif
+#ifdef HAVE_NETDB_H
+#  include <netdb.h>
+#endif
+#ifdef HAVE_WINSOCK_H
+#  include <winsock.h>
+#endif
 #include <unistd.h>
 #include <time.h>
 #include <getopt.h>
 #include <pthread.h>
+#include <signal.h>
+#include <errno.h>
 
 #include "minidom.h"
 
@@ -25,7 +42,7 @@
 #include "config.h"
 
 /* Version number */
-#define METASERV_VERSION "0.2"
+#define METASERV_VERSION "0.4"
 
 /* The directory where metaservconf.xml resides */
 #define METASERV_DIR PREFIX "/share/metaserv"
@@ -40,18 +57,71 @@
 static DOM *configuration = NULL;
 char *logfile = NULL;
 static char *cachefile = NULL;
+static int daemonmode = 0;
+static int daemonport = 0;
+static int verbosity = 0;
+pthread_mutex_t mutex;
 
-static char *metaserv_lookup(const char *class, const char *category, const char *key, int xmlformat)
+static void metaserv_cache()
+{
+	char tmp[1024];
+
+	if(cachefile)
+	{
+		logline("Caching configuration to %s", cachefile);
+		snprintf(tmp, sizeof(tmp), "%s-%li.tmp", cachefile, time(NULL));
+		unlink(tmp);
+		minidom_dumpfile(configuration, tmp);
+		rename(tmp, cachefile);
+		chmod(cachefile, S_IRUSR);
+		/*minidom_dumpfile(configuration, cachefile);*/
+	}
+}
+
+static char *xml_strdup(const char *s)
+{
+	char *s2 = NULL;
+	int i;
+	int j = 0;
+
+	if(!s) return s2;
+
+	s2 = (char*)malloc(j + 1);
+	s2[j] = 0;
+	j++;
+	for(i = 0; i < strlen(s); i++)
+	{
+		if((s[i] == '&') && ((i > strlen(s) - 5) || (strncmp(s + i, "&amp;", 5))))
+		{
+			s2 = (char*)realloc(s2, j + 5);
+			strcpy(s2 + j - 1, "&amp;");
+			s2[j + 4] = 0;
+			j += 5;
+		}
+		else
+		{
+			s2 = (char*)realloc(s2, j + 1);
+			s2[j - 1] = s[i];
+			s2[j] = 0;
+			j++;
+		}
+	}
+
+	return s2;
+}
+
+static char *metaserv_lookup(const char *class, const char *category, const char *delta, const char *key, int xmlformat)
 {
 	const char *header = "<?xml version=\"1.0\"?><resultset referer=\"query\">";
 	const char *footer = "</resultset>";
 	static char *xmlret = NULL;
-	char *ret, *r, *version;
+	char *ret, *r, *version, *prefstr;
 	int i, j, valid;
 	ELE **ele;
 	ATT **att;
 	int preference, pref;
 	char tmp[1024];
+	char *lastupdate;
 
 	if((!configuration)
 	|| (!configuration->el)
@@ -109,17 +179,30 @@ static char *metaserv_lookup(const char *class, const char *category, const char
 		r = NULL;
 		if(!strcmp(ele[i]->name, category))
 		{
-			version = MD_att(ele[i], "version");
+			version = MD_att(ele[i], "ggzmeta:version");
+			if(!version) version = MD_att(ele[i], "version"); /* backward compatibility */
 			if((!version) || (strcmp(version, key)))
 			{
 				i++;
 				continue;
 			}
 
+			if(delta)
+			{
+				lastupdate = MD_att(ele[i], "ggzmeta:update");
+				if((lastupdate) && (atoi(lastupdate) < atoi(delta)))
+				{
+					i++;
+					continue;
+				}
+			}
+
 			r = ele[i]->value;
 			if(!xmlformat)
 			{
-				pref = atoi(MD_att(ele[i], "preference"));
+				prefstr = MD_att(ele[i], "preference");
+				if(prefstr) pref = atoi(prefstr);
+				else pref = 0;
 
 				if((pref == preference) && (rand() % 2)) ret = r;
 				if((rand() % (100 + pref) > preference) || (!preference))
@@ -130,7 +213,9 @@ static char *metaserv_lookup(const char *class, const char *category, const char
 			}
 			else
 			{
-				preference = atoi(MD_att(ele[i], "preference"));
+				prefstr = MD_att(ele[i], "preference");
+				if(prefstr) preference = atoi(prefstr);
+				else preference = 0;
 				snprintf(tmp, sizeof(tmp), "<result preference=\"%i\"><uri>%s</uri>", preference, r);
 				xmlret = (char*)realloc(xmlret, strlen(xmlret) + strlen(tmp) + 1);
 				strcat(xmlret, tmp);
@@ -140,14 +225,18 @@ static char *metaserv_lookup(const char *class, const char *category, const char
 				while((att) && (att[j]))
 				{
 					if(!(strcmp(att[j]->name, "preference"))
-					|| (!strcmp(att[j]->name, "version")))
+					|| (!strncmp(att[j]->name, "ggzmeta:", 8)))
 					{
 						j++;
 						continue;
 					}
-					xmlret = (char*)realloc(xmlret, strlen(xmlret) + strlen(att[j]->name) * 2 + 5 + strlen(att[j]->value) + 1);
-					snprintf(tmp, sizeof(tmp), "<%s>%s</%s>", att[j]->name, att[j]->value, att[j]->name);
-					strcat(xmlret, tmp);
+					if(att[j]->value)
+					{
+						xmlret = (char*)realloc(xmlret, strlen(xmlret) + strlen(att[j]->name) * 2 + 5 + strlen(att[j]->value) + 1);
+						snprintf(tmp, sizeof(tmp), "<%s>%s</%s>", att[j]->name, att[j]->value, att[j]->name);
+						strcat(xmlret, tmp);
+					}
+					else logline("ALERT: key=%s without value", att[j]->name);
 					j++;
 				}
 
@@ -345,7 +434,7 @@ static char *metaserv_update(const char *class, const char *category, const char
 	char *ret;
 	char tmp[1024];
 	char *status;
-	ELE *ele, *ele2;
+	ELE *ele, *ele2, *ele3;
 	char *att2;
 	int i;
 #ifdef METASERV_OPTIMIZED
@@ -364,6 +453,7 @@ static char *metaserv_update(const char *class, const char *category, const char
 	if((!class) || (!category) || (!username) || (!password) || (!uri) || (!att) || (!atnum) || (!mode))
 	{
 		logline("Update: Missing arguments");
+logline("Update: class=%s, category=%s, username=%s, password=%s, uri=%s :: %s %i %s", class, category, username, password, uri, att, atnum, mode);
 		return NULL;
 	}
 
@@ -378,6 +468,10 @@ static char *metaserv_update(const char *class, const char *category, const char
 	{
 		status = "rejected";
 	}
+	else if(!daemonmode)
+	{
+		status = "readonly";
+	}
 	else
 	{
 		i = 0;
@@ -396,12 +490,14 @@ static char *metaserv_update(const char *class, const char *category, const char
 					j++;
 				}
 #else
-			att2 = MD_att(ele2, "class");
+				att2 = MD_att(ele2, "class");
 #endif
 			}
 			i++;
 		}
 
+		pthread_mutex_lock(&mutex);
+
 		if(att2)
 		{
 			if(!strcmp(mode, "add"))
@@ -414,7 +510,7 @@ static char *metaserv_update(const char *class, const char *category, const char
 				ele->at = att;
 				ele->elnum = 0;
 				ele->atnum = atnum;
-				ele->value = strdup(uri);
+				ele->value = xml_strdup(uri);
 
 				/* add updated data to the meta server */
 				ele2->elnum++;
@@ -425,6 +521,7 @@ static char *metaserv_update(const char *class, const char *category, const char
 				metaserv_peers(ele);
 
 				/* dump new configuration */
+				metaserv_cache();
 				/*minidom_dump(configuration);*/
 			}
 			else if(!strcmp(mode, "modify"))
@@ -433,13 +530,33 @@ static char *metaserv_update(const char *class, const char *category, const char
 			else if(!strcmp(mode, "delete"))
 			{
 				/* remove XML element */
-				free(ele2);
-				ele2 = NULL;
+				i = 0;
+				while((ele2->el) && (ele2->el[i]))
+				{
+					ele3 = ele2->el[i];
+					if(!strcmp(ele3->name, "connection"))
+					{
+						if((ele3->value) && (!strcmp(ele3->value, uri)))
+						{
+							free(ele3);
+							/*ele3 = NULL;*/
+							ele2->el[i] = NULL;
+							for(j = i + 1; ele2->el[j]; j++)
+							{
+								ele2->el[j - 1] = ele2->el[j];
+								ele2->el[j] = NULL;
+							}
 
-				metaserv_peers(NULL);
+							metaserv_peers(NULL);
 
-				/* dump new configuration */
-				/*minidom_dump(configuration);*/
+							/* dump new configuration */
+							metaserv_cache();
+							/*minidom_dump(configuration);*/
+							break;
+						}
+					}
+					i++;
+				}
 			}
 			else status = "wrong";
 		}
@@ -447,6 +564,8 @@ static char *metaserv_update(const char *class, const char *category, const char
 		{
 			status = "wrong";
 		}
+
+		pthread_mutex_unlock(&mutex);
 	}
 
 	logline("Update: status=%s", status);
@@ -477,7 +596,7 @@ static char *metaserv_xml(const char *uri)
 	DOM *query;
 	ATT **att;
 	char *ret;
-	char *class, *category;
+	char *class, *category, *delta;
 	char *username, *password;
 	char *uri2, *name;
 	char *mode;
@@ -485,13 +604,15 @@ static char *metaserv_xml(const char *uri)
 #ifdef METASERV_OPTIMIZED
 	int j;
 #endif
+	char tmp[128];
 
 	ret = NULL;
 	uri2 = NULL;
 	mode = NULL;
 	atnum = 0;
 
-	logline("XML: uri=%s", uri);
+	if(verbosity)
+		logline("XML: uri=%s", uri);
 
 	query = minidom_parse(uri);
 
@@ -507,18 +628,21 @@ static char *metaserv_xml(const char *uri)
 #ifdef METASERV_OPTIMIZED
 		class = NULL;
 		category = NULL;
+		delta = NULL;
 		i = 0;
 		while((query->el->at) && (query->el->at[i]))
 		{
 			if(!strcmp(query->el->at[i]->name, "class")) class = query->el->at[i]->value;
 			if(!strcmp(query->el->at[i]->name, "type")) category = query->el->at[i]->value;
+			if(!strcmp(query->el->at[i]->name, "delta")) delta = query->el->at[i]->value;
 			i++;
 		}
 #else
 		class = MD_att(query->el, "class");
 		category = MD_att(query->el, "type");
+		delta = MD_att(query->el, "delta");
 #endif
-		ret = metaserv_lookup(class, category, query->el->value, 1);
+		ret = metaserv_lookup(class, category, delta, query->el->value, 1);
 	}
 	else if(!strcmp(query->el->name, "update"))
 	{
@@ -575,8 +699,11 @@ static char *metaserv_xml(const char *uri)
 						atnum++;
 						att = (ATT**)realloc(att, (atnum + 1) * sizeof(ATT*));
 						att[atnum - 1] = (ATT*)malloc(sizeof(ATT));
-						att[atnum - 1]->name = strdup(name);
-						att[atnum - 1]->value = strdup(query->el->el[i]->value);
+						if(!strcmp(name, "version"))
+							att[atnum - 1]->name = strdup("ggzmeta:version");
+						else
+							att[atnum - 1]->name = xml_strdup(name);
+						att[atnum - 1]->value = xml_strdup(query->el->el[i]->value);
 						att[atnum] = NULL;
 						/*printf("Got: %s/%s\n", name, query->el->el[i]->value);*/
 					}
@@ -585,6 +712,15 @@ static char *metaserv_xml(const char *uri)
 			/*if(!strcmp(query->el->el[i]->name, "uri")) uri2 = query->el->el[i]->value;*/
 			i++;
 		}
+
+		atnum++;
+		att = (ATT**)realloc(att, (atnum + 1) * sizeof(ATT*));
+		att[atnum - 1] = (ATT*)malloc(sizeof(ATT));
+		att[atnum - 1]->name = strdup("ggzmeta:update");
+		snprintf(tmp, sizeof(tmp), "%li", time(NULL) + 7200); /* 2 hours tolerance */
+		att[atnum - 1]->value = strdup(tmp);
+		att[atnum] = NULL;
+
 		ret = metaserv_update(class, category, username, password, uri2, att, atnum, mode);
 	}
 	else if(!strcmp(query->el->name, "help"))
@@ -625,7 +761,7 @@ static char *metaserv_uri(char *requri)
 		category = strdup(token);
 		token = strtok(NULL, ":/");
 		/* 0.0.5pre */
-		if(token) ret = metaserv_lookup(class, category, token, 0);
+		if(token) ret = metaserv_lookup(class, category, NULL, token, 0);
 		free(category);
 		free(class);
 	}
@@ -638,7 +774,11 @@ static char *metaserv_uri(char *requri)
 		}
 		else logline("Unknown class type in %s", token);
 	}
-	else logline("Unknown query type in %s", requri);
+	else
+	{
+		if(verbosity)
+			logline("Unknown query type in %s", requri);
+	}
 
 	while(token) token = strtok(NULL, ":/");
 
@@ -695,35 +835,46 @@ static void metaserv_shutdown()
 	minidom_free(configuration);
 }
 
-static int metaserv_work(int fd)
+static int metaserv_work(int fd, int session)
 {
 	char buffer[1024];
 	char *result;
 	int ret;
 	FILE *stream;
 
-	logline("Enter main loop");
+	logline("[%i] Enter main loop", session);
 	while(1)
 	{
 		ret = read(fd, buffer, sizeof(buffer));
 		if(ret > 0)
 		{
 			buffer[ret - 1] = 0;
-			logline("Request: buffer=%s", buffer);
+			if(verbosity)
+				logline("[%i] Request: buffer=%s", session, buffer);
+			else
+				logline("[%i] Got request", session);
 			result = metamagic(buffer);
 
 			stream = fdopen(fd, "w");
-			if(result)
+			if(stream)
 			{
-				fprintf(stream, "%s\n", result);
-				logline("Result: result=%s", result);
+				if(result)
+				{
+					fprintf(stream, "%s\n", result);
+					if(verbosity)
+						logline("[%i] Result: result=%s", session, result);
+					else
+						logline("[%i] Sent result", session);
+				}
+				else
+				{
+					fprintf(stream, "\n");
+					logline("[%i] No result", session);
+				}
+				fflush(stream);
+				//fclose(stream);
 			}
-			else
-			{
-				fprintf(stream, "\n");
-				logline("No result");
-			}
-			fflush(stream);
+			else logline("[%i] Broken pipe", session);
 		}
 		else return 0;
 	}
@@ -739,6 +890,9 @@ static void *metaserv_worker(void *arg)
 	int ret;
 	struct hostent host, *hp;
 	int herr;
+	const char *peer;
+	static long session = 0;
+	int tmpsession;
 	char tmp[1024];
 
 	fd = *(int*)arg;
@@ -746,23 +900,15 @@ static void *metaserv_worker(void *arg)
 
 	namelen = sizeof(struct sockaddr_in);
 	getpeername(fd, (struct sockaddr*)&peername, &namelen);
-	ret = gethostbyaddr_r(&peername.sin_addr, sizeof(struct in_addr), AF_INET, &host, tmp, 1024, &hp, &herr);
-	if(!ret) snprintf(tmp, sizeof(tmp), host.h_name);
-	else snprintf(tmp, sizeof(tmp), "(unknown)");
-	logline("Accepted connection from %s", tmp);
-	metaserv_work(fd);
-	logline("Disconnection from %s", tmp);
+	ret = gethostbyaddr_r(&peername.sin_addr, sizeof(struct in_addr), AF_INET, &host, tmp, sizeof(tmp), &hp, &herr);
+	if(!ret) peer = host.h_name;
+	else peer = "(unknown)";
 
-	if(cachefile)
-	{
-		logline("Caching configuration to %s", cachefile);
-
-		snprintf(tmp, sizeof(tmp), "/tmp/metaservdump-%i-%li", fd, time(NULL));
-		unlink(tmp);
-		minidom_dumpfile(configuration, tmp);
-		rename(tmp, cachefile);
-		chmod(cachefile, S_IRUSR);
-	}
+	tmpsession = session++;
+	logline("Accepted connection from %s as [%i]", peer, tmpsession);
+	metaserv_work(fd, tmpsession);
+	logline("Disconnection from %s as [%i]", peer, tmpsession);
+	close(fd);
 
 	return NULL;
 }
@@ -774,25 +920,33 @@ static void metaserv_daemon()
 	int on;
 	int ret;
 	int fd;
-	pthread_t thread;
+	pthread_t *thread;
 	int *arg;
 
 	sock = socket(PF_INET, SOCK_STREAM, 0);
 	if(sock < 0) return;
 
 	name.sin_family = AF_INET;
-	name.sin_port = htons(METASERV_PORT);
+	name.sin_port = htons(daemonport);
 	name.sin_addr.s_addr = htonl(INADDR_ANY);
 
 	on = 1;
 	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
 
 	ret = bind(sock, (struct sockaddr*)&name, sizeof(name));
-	if(ret < 0) return;
+	if(ret < 0)
+	{
+		fprintf(stderr, "Couldn't bind to port, maybe already running?\n");
+		return;
+	}
 
 	ret = listen(sock, 1);
 	if(ret < 0) return;
 
+	pthread_mutex_init(&mutex, NULL);
+
+	signal(SIGPIPE, SIG_IGN);
+
 	while(1)
 	{
 		fd = accept(sock, NULL, NULL);
@@ -800,7 +954,10 @@ static void metaserv_daemon()
 
 		arg = (int*)malloc(sizeof(int));
 		*arg = fd;
-		pthread_create(&thread, NULL, metaserv_worker, arg);
+		thread = (pthread_t*)malloc(sizeof(pthread_t));
+		ret = pthread_create(thread, NULL, metaserv_worker, arg);
+		if(ret) close(fd);
+		else pthread_detach(*thread);
 	}
 }
 
@@ -810,7 +967,6 @@ int main(int argc, char *argv[])
 	char *config = NULL;
 	int optindex;
 	int opt;
-	int daemonmode = 0;
 	struct option options[] =
 	{
 		{"configuration", required_argument, 0, 'c'},
@@ -819,12 +975,14 @@ int main(int argc, char *argv[])
 		{"help", no_argument, 0, 'h'},
 		{"version", no_argument, 0, 'v'},
 		{"logfile", required_argument, 0, 'l'},
+		{"port", required_argument, 0, 'p'},
+		{"verbose", no_argument, 0, 'V'},
 		{0, 0, 0, 0}
 	};
 
 	while(1)
 	{
-		opt = getopt_long(argc, argv, "c:C:dhl:v", options, &optindex);
+		opt = getopt_long(argc, argv, "c:C:dhl:p:vV", options, &optindex);
 		if(opt == -1) break;
 		switch(opt)
 		{
@@ -840,14 +998,16 @@ int main(int argc, char *argv[])
 			case 'h':
 				printf("The GGZ Gaming Zone Meta Server\n");
 				printf("Version %s (GGZ %s)\n", METASERV_VERSION, VERSION);
-				printf("Copyright (C) 2001 - 2003 Josef Spillner, josef@ggzgamingzone.org\n");
+				printf("Copyright (C) 2001 - 2004 Josef Spillner, josef@ggzgamingzone.org\n");
 				printf("Published under GNU GPL conditions\n\n");
 				printf("Available options:\n");
 				printf("[-c | --configuration]: Use this configuration file\n");
 				printf("[-C | --cache        ]: Use this cache file\n");
 				printf("[-d | --daemon       ]: Run in daemon mode\n");
 				printf("[-h | --help         ]: Show this help\n");
+				printf("[-p | --port         ]: Use this port instead of %i\n", METASERV_PORT);
 				printf("[-v | --version      ]: Display version number\n");
+				printf("[-V | --verbose      ]: More verbose log file\n");
 				printf("[-l | --logfile      ]: Log events into this file\n");
 				exit(0);
 				break;
@@ -855,20 +1015,30 @@ int main(int argc, char *argv[])
 				printf("%s\n", METASERV_VERSION);
 				exit(0);
 				break;
+			case 'V':
+				verbosity = 1;
+				break;
 			case 'l':
 				logfile = optarg;
 				break;
+			case 'p':
+				daemonport = atoi(optarg);
+				break;
 			default:
 				printf("Unknown option, try --help.\n");
 				exit(-1);
 		}
 	}
 
+	if(!daemonport) daemonport = METASERV_PORT;
+
+	if((verbosity) && (!logfile)) logfile = "stdout";
+
 	metaserv_init(config);
 
 	if(!daemonmode)
 	{
-		ret = metaserv_work(STDIN_FILENO);
+		ret = metaserv_work(STDIN_FILENO, 0);
 		logline("Disconnection");
 	}
 	else metaserv_daemon();
